#import <Foundation/Foundation.h>
#import <AppKit/AppKit.h>

@interface DiagnosticsPanel : NSPanel
{
    NSTextField *textField;
    NSScrollView *scrollView;
    NSTextView *textView;

    NSSize fontDims;
    NSSize strDims;
}

- initWithTitle:(NSString *)title andContent:(NSString *)diagnostics;

- (void)appendContent:(NSString *)diagnostics;

- (void)setTitle:(NSString *)newTitle;

- done:(id)sender;
@end

NSString *RunPrompt(NSString *, NSString *);
void ShowDiagnostics(NSString *, NSString *);
